sparseflag=True
load 'quadratic_forms.sage'
load 'sdp_solver.sage'
load 'linear_algebra.sage'

### Polynomials
indeterminates=['x', 'y', 'z']
polynomial_ring=PolynomialRing(QQ, indeterminates)
  
monomial_list=[polynomial_ring('x')^a * polynomial_ring('y')^b * polynomial_ring('z')^c for a in range(4) for b in range(4) for c in range(4) if a+b+c<=3]

lhs_polynomials=[polynomial_ring('x*y + 3*y^2 + z + x^3 + 1'), polynomial_ring('x - 2*y^2 + 5*z^3 + 2'), polynomial_ring('x^2 - z + y'), 1]
lhs_polynomials.append(-(lhs_polynomials[0]+polynomial_ring('3+(x+5*y)^2')*lhs_polynomials[1]+lhs_polynomials[2])-polynomial_ring('1+x^2'));

lhs_monomial_lists=[monomial_list for p in lhs_polynomials]
rhs_polynomial=polynomial_ring(-1)

# indeterminates=['x', 'y', 'z']
# polynomial_ring=PolynomialRing(QQ, indeterminates)

# monomial_list=[polynomial_ring('x')^a * polynomial_ring('y')^b * polynomial_ring('z')^c for a in range(4) for b in range(4) for c in range(4) if a+b+c<=4]

# lhs_polynomials=[polynomial_ring(p) for p in ['1']]
# lhs_monomial_lists=[monomial_list for p in lhs_polynomials]
# assert len(lhs_polynomials) == len(lhs_monomial_lists)

# rhs_polynomial=polynomial_ring('3/4*(5*x+3*y^2+z^2)^2+2*(5*x+3*y^2+z^2)+5+(11*x+2*y+z)^4')
#rhs_polynomial=polynomial_ring('3/4*(551*x^2+3*y^2+z)^2+2*(551*x^2+3*y^2+z)+5+(23*x+13*y+z)^2')
# rhs_polynomial=polynomial_ring('3/4*(x+3*y^2+z)^2+2*(x+3*y^2+z)+5')
#rhs_polynomial=polynomial_ring('3/4*(x+3*y^2+z)^2+2*(x+3*y^2+z)+5+(1+x+y+z)^2')


def coeff_name(block, i, j):
  if i<j:
    k=i
    i=j
    j=k
  return "mat_%d_%d_%d" %(block, i, j)

coeff_positions=[(block, i, j)\
             for block in xrange(len(lhs_monomial_lists))\
             for i in xrange(len(lhs_monomial_lists[block]))\
             for j in xrange(i+1)]
coeff_positions_rev = dict([(coeff_positions[k], k) for k in xrange(len(coeff_positions))])

coeff_names=[coeff_name(block, i, j) for (block, i, j) in coeff_positions]

def get_coeff_position(block, i, j):
  if i < j:
    return coeff_positions_rev[(block, j, i)]
  else:
    return coeff_positions_rev[(block, i, j)]
  
def vector_to_blocks(v):
  blocks = [matrix(QQ, len(block), sparse=sparseflag) for block in lhs_monomial_lists]
  for k in v.nonzero_positions():
    (block, i, j)=coeff_positions[k]
    blocks[block][i,j] = blocks[block][j,i] = v[k]
  return blocks

def polynomial_from_blocks(blocks):
  total = polynomial_ring(0)
  for i in xrange(len(lhs_polynomials)):
    total += apply_qmatrix(blocks[i], vector(lhs_monomial_lists[i])) * lhs_polynomials[i]
  return total

def apply_solution(solution, zf0b, zfisb):
  assert len(solution) == len(zfisb)
  result = copy(zf0b)
  for block in xrange(len(result)):
    for i in xrange(len(solution)):
      result[block] += solution[i] * zfisb[i][block]
  return result

# Sun Oct 11 22:04:35 CEST 2009 cervin
# 0.77 s/tour + 1s lancement

constraint_space = VectorSpace(QQ, len(coeff_names)+1, sparse=sparseflag)

#def essai():
for numero_essai in range(1):
  constraint_map = {}
  for (exponents_t, scalar) in rhs_polynomial.dict().iteritems():
    exponents = tuple(exponents_t)
    v=constraint_space(0)
    v[len(v)-1]=-scalar
    constraint_map[exponents]=v

  for block in xrange(len(lhs_polynomials)):
    monomials = lhs_monomial_lists[block]
    for i in xrange(len(monomials)):
      for j in xrange(len(monomials)):
        k=get_coeff_position(block, i, j)
        for (exponents_t, coeff) in (monomials[i]*monomials[j]*lhs_polynomials[block]).dict().iteritems():
          exponents=tuple(exponents_t)
          if not constraint_map.has_key(exponents):
            constraint_map[exponents]=constraint_space(0)
          constraint_map[exponents][k] += coeff

  constraint_matrix = build_from_rows(constraint_map.values())
  print "echelonizing"
  constraint_matrix.echelonize();
    
  nonzero_rows = constraint_matrix.pivot_rows();
  if len(nonzero_rows)==0:
    assert false
  last_nonzero_row=nonzero_rows[len(nonzero_rows)-1]
  if last_nonzero_row < constraint_matrix.nrows()-1:
    constraint_matrix=constraint_matrix[0:last_nonzero_row+1,:]

  print "echelonized, computing kernel"
  print "constraints:", constraint_matrix.nrows()
  constraint_solution = right_kernel_matrix_from_echelon(constraint_matrix)

  iteration_count = 0
  has_solution=True

  while True:
    print "iteration count:", iteration_count
        
    iteration_count += 1
      
    fis=[]
    f0=()
    for row in constraint_solution.rows():
      if row[len(row)-1]==0:
        fis.append(row[0:len(row)-1])
      else:
        if f0==():
          f0=row[0:len(row)-1]/row[len(row)-1]
        else:
          assert false
          
    if f0==():
      has_solution=False
      break
    
    print "dimension of search space:", len(fis)

#    assert polynomial_from_blocks(vector_to_blocks(f0))==rhs_polynomial
#    for fi in fis:
#      assert polynomial_from_blocks(vector_to_blocks(f0+fi))==rhs_polynomial

    print 'conversion to blocks'
    f0b = vector_to_blocks(f0)
    fisb = map(vector_to_blocks, fis)

    if len(fisb)==0:
      solution=[]
      solution_q=[]
      for m in f0b:
        if not is_positive_semidefinite(m):
          has_solution=False
      break

    positive_ms = [() for i in f0b]
    f0b2 = [() for m in f0b]
    fisb2 = [ [() for m in fib] for fib in fisb]

    for block in xrange(len(f0b)):
      # vectors in kernels for all matrices are in final kernel
      print 'assembling matrix'
      matrices = [f0b[block]]
      for fib in fisb:
        matrices.append(fib[block])
      
      compound_matrix = stack_matrices(matrices)
 
      print 'computing kernel basis of', compound_matrix.nrows(), "x", compound_matrix.ncols()
      kernel_matrix = right_kernel_matrix(compound_matrix)
      
      print 'transferring to positive basis matrices'
      if kernel_matrix.nrows() > 0:
        positive_ms[block]=right_kernel_matrix(kernel_matrix)
      else:
        positive_ms[block]=identity_matrix(QQ,f0b[block].nrows(),sparse=sparseflag)

      print 'multiplying blocks'
      f0b2[block] = positive_ms[block] * f0b[block] * positive_ms[block].transpose()
      
      for i in xrange(len(fisb)):
        fisb2[i][block] = positive_ms[block] * fisb[i][block] * positive_ms[block].transpose()

    print 'solve sdp'
    solution=sdp_solve("sosproblem-%02d"%iteration_count, [-m for m in f0b2], fisb2, [0 for fib2 in fisb2])

    solution_q=vector(QQ,solution)

    print 'get new blocks, rationals'
    qblocks2=apply_solution(solution_q, f0b2, fisb2)

    print 'get new blocks, floats'
    numblocks2=apply_solution(solution, f0b2, fisb2)

    print "solving LLL problems"
    has_solution=True
    kernel_constraints = []
    for block in xrange(len(f0b)):
      (positives, zeroes, negatives) = signature(qblocks2[block])
      killer = zeroes + negatives
      if killer > 0:
        # print "eigenvalues:", numblocks2[block].eigenvalues()
        has_solution=False
        if len(solution_q)==0:
          break

        o=numblocks2[block]
        scale_factor=1e17/o.norm('frob') #tune
        olat=stack_matrices([o*scale_factor, identity_matrix(o.ncols(), sparse=sparseflag)])
        #kernel_vector2=pari(olat).qflll().python().column(0)
        red=olat.transpose().dense_matrix().apply_map(round,ZZ).LLL()[0:o.nrows(),o.nrows():olat.nrows()]

        for kernel_vector2 in red.rows()[:killer]:

          kernel_vector=positive_ms[block].transpose() * kernel_vector2
          eps=(o*kernel_vector2).norm()/n(kernel_vector2.norm())

          print "kernel vector:", kernel_vector, 'eps:', eps

          dim = f0b[block].nrows()
          for i in xrange(dim):
            constraint=constraint_space(0)
            for j in xrange(dim):
              constraint[get_coeff_position(block, i, j)] = kernel_vector[j]
            kernel_constraints.append(constraint)

    if has_solution:
      break
    
    if len(kernel_constraints) == 0:
      has_solution=False
      break
    
    print "building new constraint matrix"
    constraint_matrix = build_from_rows(kernel_constraints)
    print "rebuilt"

    re_basis=right_kernel_matrix(constraint_matrix * constraint_solution.transpose())
    constraint_solution = re_basis * constraint_solution
    

  if has_solution:
    matrices=apply_solution(solution_q, f0b, fisb)

    total=polynomial_ring(0)
    for block in xrange(len(lhs_polynomials)):
      sos=polynomial_ring(0)
      for (coeff, v) in sdp_matrix_to_sos(matrices[block]):
        sos += coeff * (v*vector(lhs_monomial_lists[block]))^2
      total += lhs_polynomials[block] * sos

    print "SOLUTION IS CORRECT:", total == rhs_polynomial


  else:
    print "NO SOLUTION"

#essai()

#import cProfile
#cProfile.run('essai()', 'profiling/profile6.dat')

